from __future__ import absolute_import
import random
import torch
from .utils import batchify
from torch.autograd import Variable


def get_next_word_batch(source, i, seq_len=35, evaluation=False, return_sequence=True):
    seq_len = min(seq_len, len(source) - 1 - i)
    data = Variable(source[i:i + seq_len], volatile=evaluation).cuda()
    if return_sequence:
        target = Variable(source[i + 1:i + 1 + seq_len].view(-1)).cuda()
    else:
        target = Variable(source[i + seq_len].view(-1)).cuda()

    return data, target


class Dictionary(object):
    def __init__(self):
        self.word2idx = {}
        self.idx2word = []
        self.null_token = '<NULL>'

    def encode_word2idx(self, word, enable_null=False):
        if word in self.word2idx:
            return self.word2idx[word]
        elif enable_null:
            return self.word2idx[self.null_token]
        else:
            raise IndexError('Word not in dictionary, if want to replace with Null Token call enable_null=True')

    def encode_idx2word(self, idx):
        return self.idx2word[idx]

    def add_word(self, word):
        if word not in self.word2idx:
            self.idx2word.append(word)
            self.word2idx[word] = len(self.idx2word) - 1
        return self.word2idx[word]

    def __len__(self):
        return len(self.idx2word)


class simple_generator(object):
    def __init__(self, x, y, x_test=None, y_test=None, batch_size=128, input_categorical=True, output_categorical=True,
                 train_test_split=0.9):
        self.batch_size = batch_size
        self.train_test_split = train_test_split
        self.x_train, self.x_test = self.init_tensor(x, x_test, input_categorical)
        self.y_train, self.y_test = self.init_tensor(y, y_test, output_categorical)
        self.bound = len(self.y_train)

    def init_tensor(self, train_tensor, test_tensor, categorical):
        if test_tensor is None:
            total_len = len(train_tensor)
            split_point = int(total_len * self.train_test_split)
            test_tensor = train_tensor[split_point:]
            train_tensor = train_tensor[:split_point]
        if categorical:
            train_tensor = Variable(torch.cuda.LongTensor(train_tensor))
            test_tensor = Variable(torch.cuda.LongTensor(test_tensor))
        else:
            train_tensor = Variable(torch.cuda.FloatTensor(train_tensor))
            test_tensor = Variable(torch.cuda.FloatTensor(test_tensor))

        train_tensor = batchify(train_tensor, self.batch_size)
        test_tensor = batchify(test_tensor, self.batch_size)
        return train_tensor, test_tensor

    def generate(self, evaluate=False):
        # i = 0
        while 1:
            i = random.randrange(0, self.bound)
            if not evaluate:
                # Training Case
                yield self.x_train[i], self.y_train[i]
            else:
                # Evaluation Case
                yield self.x_test[i], self.y_test[i]
                # i += 1

    def __len__(self):
        return len(self.y_train)


class Onehot_Encoder(object):
    def __init__(self, possible_inputs):
        self.possible_inputs = possible_inputs
        self.word2idx = dict((w, i) for i, w in enumerate(possible_inputs))
        self.idx2word = dict((i, w) for i, w in enumerate(possible_inputs))

    def encode(self, input):
        res = Variable(torch.cuda.FloatTensor(len(input), len(self.possible_inputs)).zero_())

        for i, value in enumerate(input):
            res[i, self.word2idx[value]] = 1
        return res

    def encode_batch(self, batch_input):

        for i, b in enumerate(batch_input):
            batch_input[i] = self.encode(b)

        return torch.stack(batch_input, dim=1)

    def forward(self, input):
        return self.encode_batch(input)



from __future__ import absolute_import

import torch.nn as nn
from torch.nn import init
import numpy as np
import torch
from .base import Layer

class Embedding(Layer):
    def __init__(self, insize, outsize, name=None, pretrain=None, dictionary=None):
        super(Embedding, self).__init__([insize], [insize, outsize], name)
        self.kernel = nn.Embedding(insize, outsize)
        init.xavier_uniform(self.kernel.weight)

        if pretrain is not None:
            assert(dictionary is not None)
            print("Loading pretrained embedding from {}".format(pretrain))
            embeddings_index = {}
            with open(pretrain) as f:
                for line in f:
                    values = line.split()
                    word = values[0]
                    coefs = np.asarray(values[1:], dtype='float32')
                    embeddings_index[word] = coefs
            emb_layer = torch.cuda.FloatTensor(insize, outsize)
            
            c = 0
            for w in dictionary.word2idx.keys():
                idx = dictionary.word2idx[w]
                if w in embeddings_index:
                    cur_emb = embeddings_index[w]
                else:
                    c += 1
                    continue
                cur_emb = torch.cuda.LongTensor(cur_emb)
                emb_layer[idx, :] = cur_emb
            print("{} out of {} words not in Embedding".format(c, insize))
            self.kernel.weight = nn.Parameter(emb_layer)


from __future__ import absolute_import

import torch.nn as nn
import torch.nn.functional as F
from .base import Layer

class Softmax(Layer):
    def __init__(self, insize, outsize):
        super(Softmax, self).__init__(insize, outsize)
        self.kernel = nn.LogSoftmax(dim=2)


class Sigmoid(Layer):
    def __init__(self, insize, outsize):
        super(Sigmoid, self).__init__(insize, outsize)
        self.kernel = nn.LogSigmoid(dim=2)

        
class Activation(Layer):
    def __init__(self, non_linearity_type, name=None):
        super(Activation, self).__init__([None], [None], name)
        self.kernel = getattr(F, non_linearity_type)


from __future__ import absolute_import
from ..train import *
from ..utils import *
from ..metrics import *

class Model(object):
    def __init__(self):
        self.model = nn.Sequential()
        self.layer_num = 0
        self.complied = False
        self.data_loader = None
        self.data_generator = None
        self.training_length = None
        self.optimizer = None
        self.loss_func = None
        self.clip_norm = None
        self.progress_bar = None
        self.metrics = None
        self.output_categorical = True

    def add(self, layer):
        layer = layer.cuda()
        self.model.add_module("layer_{}".format(self.layer_num), layer)
        self.layer_num += 1

    def compile(self, optimizer="adam", criterion="crossentropy", clip_norm=True, metrics=[]):

        optimizer = get_optimizer(optimizer, self.model.parameters())
        self.criterion_string = criterion
        if isinstance(criterion, str):
            loss_func, self.output_categorical = get_loss_func(criterion)
        else:
            loss_func, self.output_categorical = criterion

        self.complied = True
        self.optimizer = optimizer
        self.loss_func = loss_func
        self.clip_norm = clip_norm
        self.metrics = metrics

    def train_one_epoch(self):
        if not self.complied:
            raise TypeError("Model need to be compiled before training")

        self.model.train()
        for batch, i in enumerate(range(0, self.training_length)):
            data, targets = next(self.data_generator)
            self.model.zero_grad()
            self.optimizer.zero_grad()
            output = self.model(data)
            loss = self.loss_func(output, targets)
            loss.backward()

            if self.clip_norm:
                torch.nn.utils.clip_grad_norm(self.model.parameters(), 0.25)

            self.optimizer.step()
            report = []

            if i % 10 == 0:
                if "loss" in self.metrics:
                    report.append(("loss", loss.data[0]))
                if "acc" in self.metrics:
                    accuracy = get_accuracy(self.criterion_string, output, targets, self.batch_size)
                    report.append(("acc", accuracy))
                if "perplexity" in self.metrics:
                    report.append(("perplexity", torch.exp(loss).data[0]))

                self.progress_bar.update(i, report)

    def repackage_hiddens(self):
        for layer in self.model:
            if layer.hidden is not None:
                layer.hidden = repackage_hidden(layer.hidden)

    def fit_generator(self, data_loader, batch_size, epochs, train_length=None):
        self.init_model(batch_size)

        if train_length is None:
            self.training_length = len(data_loader)
        else:
            self.training_length = train_length

        self.data_loader = data_loader
        self.data_generator = data_loader.generate()
        self.train_run(epochs=epochs)

    def train_run(self, epochs=1):
        try:
            for epoch in range(1, epochs + 1):
                self.progress_bar = Progbar(self.training_length)
                epoch_start_time = time.time()
                self.train_one_epoch()
                delta_time = time.time() - epoch_start_time
                print("\n")

        except KeyboardInterrupt:
            print("\n")
            print('-' * 89)
            print('Exiting from training early')

    def evaluation(self, test_loader, batch_size=1):
        self.init_model(batch_size)
        self.model.eval()

        text_generator = test_loader.generate()
        test_length = len(test_loader)
        test_progress_bar = Progbar(test_length)
        total_loss = 0
        total_acc = 0
        total_preplexity = 0

        for i in range(0, test_length):
            data, targets = next(text_generator)
            output = self.model(data)
            loss = self.loss_func(output, targets)
            total_loss += loss.data[0]
            report = []
            if "loss" in self.metrics:
                report.append(("loss", loss.data[0]))
                total_loss += loss.data[0]
            if "acc" in self.metrics:
                accuracy = get_accuracy(self.criterion_string, output, targets, self.batch_size)
                report.append(("acc", accuracy))
                total_acc += accuracy
            if "perplexity" in self.metrics:
                report.append(("perplexity", torch.exp(loss).data[0]))
                total_preplexity += total_preplexity.data[0]

            test_progress_bar.update(i, report)

        return total_loss / test_length, total_acc / test_length, total_preplexity / test_length

    def init_model(self, batch_size):
        self.batch_size = batch_size
        for layer in self.model:
            layer.batch_size = batch_size
            if layer.require_hidden:
                layer.init_hidden()

    def eval(self):
        self.init_model(1)

    def save(self, file_name):
        with open(file_name, 'wb') as f:
            torch.save(self.model, f)


class ModelGenerate(object):
    def __init__(self, model, dictionary):
        model.eval()
        self.model = model.model
        self.model.eval()
        self.dictionary = dictionary

    def generate(self, tempeture=1, num_word=1000):
        ntokens = len(self.dictionary)
        output_words = []
        input = Variable(torch.rand(1, 1).mul(ntokens).long(), volatile=True).cuda()
        for i in range(num_word):
            output = self.model(input)
            word_weights = output.squeeze().data.div(tempeture).exp().cpu()
            word_idx = torch.multinomial(word_weights, 1)[0]
            input.data.fill_(word_idx)
            word = self.dictionary.idx2word[word_idx]
            print(word, end=" ")
            output_words.append(word)

        return " ".join(output_words)


def load_model(file_name):
    with open(file_name, 'rb') as f:
        core_model = torch.load(f)

    model = Model()
    model.model = core_model
    return model


def epochs_report(epoch, delta_time, loss):
    print('-' * 89)
    print('| end of epoch {:3d} | time: {:5.2f}s | valid loss {:5.2f} |'.format(epoch, delta_time, loss))
    print('-' * 89)



import torch.nn as nn
from torch import optim


def get_loss_func(loss_string):
    
    if not isinstance(loss_string, str):
        if len(loss_string) != 2:
            raise TypeError("Customized loss layer need to return whether output is categorical")
        return loss_string
        
    loss_string = loss_string.lower()
    criterion_ls = {"crossentropy": nn.NLLLoss(), "mse": nn.MSELoss(), "binarycrossentropy": nn.BCELoss()}
    out_put_categorical = {"crossentropy": True, "mse": False, "binarycrossentropy": True}
    return criterion_ls[loss_string], out_put_categorical[loss_string]


def get_accuracy(loss_string, output, targets, batch_size, threshold=0.5):
    if loss_string == "crossentropy":
        max_index = output.max(dim=1)[1]
        correct = (max_index == targets).sum()
        accuracy = int(correct.data) / len(targets)

    elif loss_string == "binarycrossentropy":
        binary_output = output > threshold
        correct = sum((binary_output.float() == targets).data.cpu().numpy()[0])
        accuracy = correct / batch_size

    return accuracy


def get_optimizer(optimizer_string, parameters):
    if not isinstance(optimizer_string, str):
        return optimizer_string
    else:
        opt_ls = {"adam": optim.Adam(parameters, lr=1e-3)}
        return opt_ls[optimizer_string]



from __future__ import absolute_import

import torch.nn as nn
from torch.autograd import Variable
from .base import *

class GRU(Layer):
    def __init__(self, insize, hidden_size, name=None, nlayers=1, recurrent_dropout=0.1, return_sequence=True, extract_hidden=False):
        self.insize = insize
        if return_sequence:
            inshape = [None, insize]
            outshape = [None, hidden_size]
        else:
            inshape = [insize]
            outshape = [hidden_size]
        
        super(GRU, self).__init__(inshape, outshape, name)
        self.hidden_size = hidden_size
        self.extract_hidden = extract_hidden
        self.nlayers = nlayers
        self.recurrent_dropout = recurrent_dropout
        self.kernel = nn.GRU(insize, hidden_size, self.nlayers, dropout=self.recurrent_dropout)
        self.return_sequence = return_sequence
        self.require_hidden = True

    def forward(self, input):
        input = input.view(-1, self.batch_size, self.insize)
        self.hidden = self.init_hidden()
        output, hidden = self.kernel(input, self.hidden)

        if not self.return_sequence:
            output = output[-1:]

        if self.extract_hidden:
            return hidden
        else:
            return output.view(self.batch_size, -1, self.hidden_size)

    def init_hidden(self):
        weight = next(self.parameters()).data
        self.hidden = Variable(weight.new(self.nlayers, self.batch_size, self.hidden_size).zero_().cuda())



from __future__ import absolute_import

import torch.nn as nn
from torch.autograd import Variable
from .base import Layer

class Input(Layer):
    def __init__(self, insize, name=None):
        super(Input, self).__init__([insize], [insize], name)
        
    def forward(self, input):
        return input
    
    
class Dropout(Layer):
    def __init__(self, dropout, name=None):
        super(Dropout, self).__init__([None], [None], name)
        self.kernel = nn.Dropout(dropout)
        
class Reshape(Layer):
    def __init__(self, shape, name=None):
        super(Reshape, self).__init__([None], [None], name)
        self.shape = shape
    def forward(self, input):
        return input.view(*self.shape)


from __future__ import absolute_import


class Model(nn.Module):
    def __init__(self, input_layer, output_layer):
        super(Model, self).__init__()
        self.input_layer = input_layer
        self.output_layer = output_layer
        self.layers = []
        self.params = []
        self.traverse_layers(input_layer)
        
        
    def forward_layer(self, input, cur_layer):
        output = cur_layer.forward(input)
        if not cur_layer.out_bound_layers:
            return output
        
        for layer in cur_layer.out_bound_layers:
            return self.forward_layer(output, layer)
    
    def parameters(self):
        return self.params
    
    def traverse_layers(self, cur_layer):
        for layer in cur_layer.out_bound_layers:
            self.layers.append(layer)
            self.params += list(layer.parameters())
            self.traverse_layers(layer)
            
    def change_batch_size(self, bsz):
        self.batch_size = batch_size

        for layer in self.layers:
            layer.cuda()
            layer.update_batch_size(bsz)
            
            
    def forward(self, input):
        return self.forward_layer(input, self.input_layer)
    
    
    def compile(self, optimizer="adam", criterion="crossentropy", clip_norm=True, metrics=[]):

        optimizer = get_optimizer(optimizer, self.parameters())
        loss_func, self.output_categorical = get_loss_func(criterion)
        self.criterion_string = criterion
        self.complied = True
        self.optimizer = optimizer
        self.loss_func = loss_func
        self.clip_norm = clip_norm
        self.metrics = metrics

    def train_one_epoch(self):
        if not self.complied:
            raise TypeError("Model need to be compiled before training")

        self.train()
        for batch, i in enumerate(range(0, self.training_length)):
            data, targets = next(self.data_generator)
            self.zero_grad()
            self.optimizer.zero_grad()
            output = self(data)
            loss = self.loss_func(output, targets)
            loss.backward()

            if self.clip_norm:
                torch.nn.utils.clip_grad_norm(self.parameters(), 0.25)

            self.optimizer.step()
            report = []

            if i % 20 == 0:
                if "loss" in self.metrics:
                    report.append(("loss", loss.data[0]))
                if "acc" in self.metrics:
                    accuracy = get_accuracy(self.criterion_string, output, targets, self.batch_size)
                    report.append(("acc", accuracy))
                if "perplexity" in self.metrics:
                    report.append(("perplexity", torch.exp(loss).data[0]))

                self.progress_bar.update(i, report)
                
    def fit_generator(self, data_loader, batch_size, epochs, train_length=None):
        self.change_batch_size(batch_size)

        if train_length is None:
            self.training_length = len(data_loader)
        else:
            self.training_length = train_length

        self.data_loader = data_loader
        self.data_generator = data_loader.generate()
        self.train_run(epochs=epochs)

    def train_run(self, epochs=1):
        try:
            for epoch in range(1, epochs + 1):
                self.progress_bar = Progbar(self.training_length)
                epoch_start_time = time.time()
                self.train_one_epoch()
                delta_time = time.time() - epoch_start_time
                print("\n")

        except KeyboardInterrupt:
            print("\n")
            print('-' * 89)
            print('Exiting from training early')

    def evaluation(self, test_loader, batch_size=1):
        self.change_batch_size(batch_size)
        self.eval()

        text_generator = test_loader.generate()
        test_length = len(test_loader)
        test_progress_bar = Progbar(test_length)
        total_loss = 0
        total_acc = 0
        total_preplexity = 0

        for i in range(0, test_length):
            data, targets = next(text_generator)
            output = self(data)
            loss = self.loss_func(output, targets)
            total_loss += loss.data[0]
            report = []
            if "loss" in self.metrics:
                report.append(("loss", loss.data[0]))
                total_loss += loss.data[0]
            if "acc" in self.metrics:
                accuracy = get_accuracy(self.criterion_string, output, targets, self.batch_size)
                report.append(("acc", accuracy))
                total_acc += accuracy
            if "perplexity" in self.metrics:
                report.append(("perplexity", torch.exp(loss).data[0]))
                total_preplexity += total_preplexity.data[0]
            
            if i % 20 == 0:
                test_progress_bar.update(i, report)

        return total_loss / test_length, total_acc / test_length, total_preplexity / test_length


    def save(self, file_name):
        with open(file_name, 'wb') as f:
            torch.save(self, f)
    
    


from __future__ import absolute_import

import torch.nn as nn


def create_layer_name(layer, idx):
    name = layer.__class__.__name__
    return name + "_" + str(idx)
    

class Layer(nn.Module):
    def __init__(self, inshape, outshape, name):
        super(Layer, self).__init__()
        
        self.batch_size = None
        self.inshape = [self.batch_size, *inshape]
        self.outshape = [self.batch_size, *outshape]
        self.name = name
        self.in_bound_layers = []
        self.out_bound_layers = []
        
        self.kernel = None
        self.require_hidden = False
        self.hidden = None

    def forward(self, input):
        return self.kernel(input)
    
    def update_batch_size(self, bsz):
        self.batch_size = bsz
        self.inshape[0] = bsz
        self.outshape[0] = bsz

    def __call__(self, layer):
        self.in_bound_layers.append(layer)
        layer.out_bound_layers.append(self)
        return self


from __future__ import absolute_import

import unicodedata
import re
import os
from torch.autograd import Variable


def unicode_to_ascii(s):
    return ''.join(
        c for c in unicodedata.normalize('NFD', s)
        if unicodedata.category(c) != 'Mn'
    )


def normalize_string(s):
    s = unicode_to_ascii(s.lower().strip())
    s = re.sub(r"(')", r"", s)
    s = re.sub(r"[0-9]+", "NUM", s)
    s = re.sub(r"([.!?,:-;\"])", r" \1", s)
    s = re.sub(r"[^a-zA-Z.!?,-:;\"]+", r" ", s)
    return s


def fix_review(review):
    sent = normalize_string(review)
    sep = re.findall(r'([a-z][^\.!?]*[\.!? ]*)', sent)
    sep = ["SOS " + s + " EOS" for s in sep]
    return "SOP " + " ".join(sep) + " EOP"


def init_gpu(gpu_num):
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_num)


def batchify(data, bsz):
    nbatch = data.size(0) // bsz
    data = data.narrow(0, 0, nbatch * bsz)
    data = data.view(bsz, -1).t().contiguous()
    return data


def repackage_hidden(h):
    if type(h) == Variable:
        return Variable(h.data)
    else:
        return tuple(repackage_hidden(v) for v in h)



from __future__ import absolute_import

import torch.nn as nn
from torch.nn import init
from .base import Layer

class Linear(Layer):
    def __init__(self, insize, outsize, name=None):
        super(Linear, self).__init__([insize], [outsize], name)
        self.out_shape = outsize
        self.kernel = nn.Linear(insize, self.out_shape)
        self.in_shape = insize
        init.xavier_uniform(self.kernel.weight)


from setuptools import setup
from setuptools import find_packages

setup(name='eru',
      version='0.0.1',
      description='Deep Learning for all',
      author='Shawn Shan',
      author_email='shansixioing@uchicago.edu',
      url='https://github.com/Shawn-Shan/eru',
      license='MIT',
      install_requires=[],

      classifiers=[
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      packages=find_packages())



from __future__ import absolute_import
from .utils import *
from .data import *
from collections import Counter
import pickle


class MutiClassSequensial_Corpus(object):
    def __init__(self, path_dictionary=[], load_dictionary=False, keep_rate=0.995):
        self.dictionary = Dictionary()
        self.path_dictionary = path_dictionary
        self.keep_rate = keep_rate
        if not load_dictionary:
            self.update_dictionary()
        else:
            print("Load Dictionary")
            if isinstance(load_dictionary, str):
                corpus = pickle.load(open(load_dictionary, "rb"))
                self.dictionary = corpus.dictionary
            else:
                self.dictionary = load_dictionary

        self.content = {}

        for cur_dict in self.path_dictionary:
            sub_dictionary, label = cur_dict
            self.content[label] = {}
            cur_train_txt = os.path.join(sub_dictionary, 'train.txt')
            cur_test_txt = os.path.join(sub_dictionary, 'test.txt')

            if os.path.exists(cur_train_txt):
                cur_train = self.tokenize(cur_train_txt)
                self.content[label]["train"] = cur_train

            if os.path.exists(cur_test_txt):
                cur_test = self.tokenize(cur_test_txt)
                self.content[label]["test"] = cur_test

    def append_sub_corpus(self, path):
        with open(os.path.join(path, 'train.txt'), 'r') as f:
            for line in f:
                words = line.split()
                for word in words:
                    if word in self.total_count:
                        self.total_count[word] += 1
                    else:
                        self.total_count[word] = 1

    def update_dictionary(self, addition_words=[]):
        self.total_count = {}

        # Add words to the dictionary
        for cur_dict in self.path_dictionary:
            sub_dictionary, label = cur_dict
            print(sub_dictionary)
            self.append_sub_corpus(sub_dictionary)

        self.total_count = Counter(self.total_count)
        total_len = sum(self.total_count.values())
        for max_count in range(0, 1000000, 1000):
            new_dict = dict(self.total_count.most_common(max_count))
            presentage = sum(new_dict.values()) / total_len
            if presentage > self.keep_rate:
                print("Total {} words, Keep top {} words".format(str(len(self.total_count)), str(max_count)))
                break

        # Add to dictionary
        for key in new_dict.keys():
            self.dictionary.add_word(key)

        for ele in addition_words:
            self.dictionary.add_word(ele)

        self.dictionary.add_word('<NULL>')

    def tokenize(self, path):
        """Tokenizes a text file."""
        with open(path, 'r') as f:
            tokens = 0
            for line in f:
                words = line.split()
                tokens += len(words)

        with open(path, 'r') as f:
            ids = torch.LongTensor(tokens)
            token = 0
            for line in f:
                words = line.split()
                for word in words:
                    if word not in self.dictionary.word2idx:
                        word = "<NULL>"
                    ids[token] = self.dictionary.word2idx[word]
                    token += 1

        return ids


class MutiClassSequensial_Loader(object):
    def __init__(self, corpus, batch_size, evaluation=False, seq_len=35, output_categorical=True):
        self.batch_size = batch_size
        self.evaluation = evaluation
        self.seq_len = seq_len
        self.output_categorical = output_categorical
        self.corpus = corpus
        self.data = self.corpus.content
        self.data_dictionary = {}
        index = 0
        for key, value in self.data.items():
            label = key
            if evaluation:
                data = self.data[label]['test']
            else:
                data = self.data[label]['train']

            data = batchify(data, self.batch_size)

            self.data_dictionary[index] = {}
            self.data_dictionary[index]["label"] = label
            self.data_dictionary[index]["data"] = data
            self.data_dictionary[index]["bound"] = len(data) - 1 - self.seq_len
            index += 1
        self.index = index

    def generate(self):
        if self.evaluation:
            index = 0
            i = 0
            while 1:
                select = self.data_dictionary[index]
                source = select["data"]
                label = select["label"]
                bound = select["bound"]
                while i < bound:
                    batch_x = Variable(source[i:i + self.seq_len]).cuda()
                    if self.output_categorical:
                        target = Variable(torch.cuda.LongTensor([label] * self.batch_size)).view(1, -1)
                    else:
                        target = Variable(torch.cuda.FloatTensor([label] * self.batch_size)).view(1, -1)
                    yield batch_x, target
                    i += 1
                index += 1
                i = 0
        else:
            while 1:
                pointer = random.choice(range(self.index))
                selected = self.data_dictionary[pointer]
                i = random.randrange(0, selected["bound"])
                source = selected["data"]
                label = selected["label"]
                batch_x = Variable(source[i:i + self.seq_len]).cuda()
                if self.output_categorical:
                    target = Variable(torch.cuda.LongTensor([label] * self.batch_size)).view(1, -1)
                else:
                    target = Variable(torch.cuda.FloatTensor([label] * self.batch_size)).view(1, -1)

                yield batch_x, target

    def __len__(self):
        total_len = 0
        for key, value in self.data_dictionary.items():
            bound = value["bound"]
            total_len += bound
        return total_len
